#include "common.h"

namespace common
{
    int read_pheno_and_biomarker( const std::string & filename, 
                          std::vector<measurement> & biomarker_measurements,
                          VectorXd & Y_pheno_vec,
                          MatrixXd & covar_mxd,
                          std::vector<subject_data_type> & subject_data_vec,
                          std::vector< std::string > & covar_name_vec,
                          bool & isPhenoBinary )
    {
        std::ifstream in_file( filename );            
        if (!in_file.is_open())
        {            
            std::cout << "read_pheno_data : ";
            std::cout << "error : could not open " << filename << std::endl;
            return FILE_ERROR;
        }
        std::cout << "Reading input from " << filename << std::endl;
        std::string line;
        size_t num_X_cols = 0;
        size_t num_rows = 0;
        bool labels_are_read = false;
        size_t full_data_count = 0;
        isPhenoBinary = true;
        while (getline(in_file,line))
        {
            if ( ! labels_are_read )
            { //first line
                //covar_name_vec.push_back( "Constant" );
                labels_are_read = true;
                std::string buff;
                std::stringstream ss(line);
                size_t col_index = 0;
                const size_t FIRST_COVAR_COL = 3; 
                //first col has index of 0 so FIRST_COVAR_COL corresponds 
                //to the fourth col
                
                while (ss >> buff )
                {
                   if ( col_index >= FIRST_COVAR_COL )
                   {
                       covar_name_vec.push_back( buff );
                   }
                   ++col_index;
                }
                continue;
            } //first line
            std::string buff;
            std::stringstream ss(line);
            std::vector <double> line_elements;
            std::vector <std::string> line_strings;
            while (ss >> buff)
            {
               line_strings.push_back( buff );
               double element = std::stod(buff);
               line_elements.push_back(element);
               //NA in covariates is replaced with zero
            }
            if ( line_strings[0].compare( "NA" ) )
            {
                subject_data_vec.push_back( FULL_DATA );
                ++full_data_count;
            }
            else
            {
                subject_data_vec.push_back( MISSING_DATA );
            }
            if (line_elements.size() > 0)
            {
                num_rows++;
                if (num_X_cols == 0)
                {
                    //first time
                    num_X_cols = line_elements.size() - 3;  //no space for constant
                    covar_mxd.resize(num_rows, num_X_cols);
                    Y_pheno_vec.resize(num_rows );
                }
                else
                {
                    covar_mxd.conservativeResize(num_rows, Eigen::NoChange);
                    Y_pheno_vec.conservativeResize(num_rows );
                }
                Y_pheno_vec(num_rows - 1) = line_elements[ 0 ];
                if ( isPhenoBinary )
                {
                    const int rounded_Y = round( line_elements[ 0 ] );
                    if ( rounded_Y != line_elements[ 0 ]  )
                    {
                        isPhenoBinary = false;
                    }
                    else
                    if ( ( rounded_Y != 0 ) && ( rounded_Y != 1 ) )
                    {
                       isPhenoBinary = false;
                    }
                } // if ( isPhenoBinary )
                //cout << "line_elements[0] == " << line_elements[ 0 ] << endl;
                for (size_t index = 3; index < line_elements.size(); index++)
                {
                    covar_mxd( num_rows - 1, index - 3) = line_elements[index];
                }
                measurement my_measurement;
                my_measurement.type = measure_type( round(line_elements[2]) );
                switch( my_measurement.type )
                {
                    case MEASURE_TYPE_KNOWN:
                        my_measurement.measured_value = 
                        my_measurement.lower_bound = 
                        my_measurement.upper_bound =
                        line_elements[1];
                        break;
                    case MEASURE_TYPE_BELOW_DETECT_LIMIT:
                        my_measurement.measured_value =               
                        my_measurement.lower_bound =               
                        my_measurement.upper_bound =              
                        line_elements[1];
                        break;
                    case MEASURE_TYPE_ABOVE_DETECT_LIMIT:
                        my_measurement.measured_value =
                        my_measurement.lower_bound =
                        my_measurement.upper_bound =
                        line_elements[1];
                        break;
                    case MEASURE_TYPE_MISSING_DATA:
                        my_measurement.measured_value = line_elements[1]; //does not matter
                        my_measurement.lower_bound = DBL_MIN;
                        my_measurement.upper_bound = DBL_MAX;
                        break;
                    default:
                        std::cout << "Invalid measurement type: " << my_measurement.type <<
                            std::endl;
                        return FILE_ERROR;
                } // switch 
                biomarker_measurements.push_back( my_measurement );
            } //if line_elements.size() ...
        }   //while
        in_file.close();
      
       covar_mxd.transposeInPlace();
       VectorXd full_Y_vec = VectorXd::Zero( full_data_count );
       for ( size_t index = 0, full_data_index = 0;
                     index < subject_data_vec.size(); ++index )
        {
            if ( subject_data_vec[ index ] == MISSING_DATA )
            {
                continue;
            }
            full_Y_vec( full_data_index ) = Y_pheno_vec( index );
            ++full_data_index;
        }
        Y_pheno_vec = full_Y_vec;
        //now remove subjects from covar_mxd that have missing
        //phenotype data
        MatrixXd full_X_mat = MatrixXd::Zero( covar_mxd.rows(), full_data_count );
        for ( size_t index = 0, full_data_index = 0;
                     index < subject_data_vec.size(); ++index )
        {
            if ( subject_data_vec[ index ] == MISSING_DATA )
            {
                continue;
            }
            full_X_mat.block( 0, full_data_index, covar_mxd.rows(), 1 ) =
                covar_mxd.block( 0, index, covar_mxd.rows(), 1 );
            ++full_data_index;
        }
        covar_mxd = full_X_mat;
    
        //now remove subjects from biomarker_measurements that have missing phenotype
        std::vector<measurement> biomarkers_for_subjects_with_phenotype;
        for ( size_t index = 0;
                     index < biomarker_measurements.size(); ++index )
        {
            if ( subject_data_vec[ index ] == MISSING_DATA )
            { //phenotype is missing
               continue;
            }
            const measurement my_measure = biomarker_measurements[ index ];
            biomarkers_for_subjects_with_phenotype.push_back( my_measure ); 
        } //for
        biomarker_measurements = biomarkers_for_subjects_with_phenotype;
       return NO_ERROR;
    } //method

    int read_configfile( const std::string & config_filename,
                               std::string & geno_filename,
                               std::string & pheno_filename,
                               bool & pheno_covar_snp,
                               size_t & out_freq,
                               double & min_call_rate,
                               size_t & max_iter,
                               double & converged_ratio,
                               std::string & out_suffix,
                               bool & write_covariance,
                               bool & write_full_estimate )
     {
          std::ifstream in_file( config_filename );
          if (!in_file.is_open())
          {
              std::cout << "error: read_configfile() : ";
              std::cout << "could not open " << config_filename << std::endl;
              return FILE_ERROR;
          }
          const std::string DEF_OUT_SUFFIX = ".out.txt";
          out_suffix = DEF_OUT_SUFFIX;
          const size_t DEF_MAX_ITER=10000;
          const double DEF_CONV_RATIO = 1e-4;
          max_iter = DEF_MAX_ITER;
          converged_ratio = DEF_CONV_RATIO;
          const double DEF_MIN_CALL_RATE = 0.90;
          min_call_rate = DEF_MIN_CALL_RATE;
          const size_t DEF_OUT_FREQ = 1000;
          out_freq = DEF_OUT_FREQ;
          const bool DEF_WRITE_COVARIANCE=false;
          write_covariance = DEF_WRITE_COVARIANCE;
          const bool DEF_WRITE_FULL_ESTIMATE = false;
          write_full_estimate = DEF_WRITE_FULL_ESTIMATE;
          bool found_geno_filename = false;
          bool found_pheno_filename = false;
          bool found_pheno_covar_snp = false;
          std::cout << "Reading configuration file " << config_filename << std::endl;
          std::string line;
          while (getline(in_file,line))
          {
              std::string buff;
              std::stringstream ss(line);
              std::vector<std::string> line_elements;
              while (ss >> buff)
              {
                  line_elements.push_back(buff);
              }
              const size_t num_elements = line_elements.size();
              if ( 0 == num_elements )
              {
                 //blank line
                 continue;
              }
              if ( 2 != num_elements )
              {
                 in_file.close();
                 std::cout << "read_configfile() : line in " << config_filename <<
                         "has " << num_elements << "columns.  it should have either 0 or 2"
                         << std::endl;
                 return FILE_ERROR;
              }
              if ( ( line_elements[ 0 ] != "geno_filename" ) &&
                    ( line_elements[ 0 ] != "pheno_filename" ) &&
                   ( line_elements[ 0 ] != "pheno_covar_snp" ) &&
                   ( line_elements[ 0 ] != "out_freq" ) &&
                   ( line_elements[ 0 ] != "min_call_rate" ) && 
                   ( line_elements[ 0 ] != "out_suffix" ) && 
                   ( line_elements[ 0 ] != "write_covariance" ) &&
                   ( line_elements[ 0 ] != "write_full_estimate" ) ) 
              {
                 in_file.close();
                 std::cout << "error: read_configfile(): Invalid string " << line_elements[ 0 ]
                      << " found in first column of " << config_filename << std::endl;
                 return FILE_ERROR;
              }
              if ( "out_suffix" == line_elements[ 0 ] )
              {
                 out_suffix = line_elements[ 1 ];
              }
              else
              if ( "geno_filename" == line_elements[ 0 ] )
              {
                 geno_filename = line_elements[ 1 ];
                 found_geno_filename = true;
              }
            else
             if ( "pheno_filename" == line_elements[ 0 ] )     
              {
                 pheno_filename = line_elements[ 1 ];    
                 found_pheno_filename = true;        
              }
              else
              if ( "pheno_covar_snp" == line_elements[ 0 ] )
              {
                 if ( "true" == line_elements[ 1 ] )
                 {
                    pheno_covar_snp = true;
                 }
                 else
                 if ( "false" == line_elements[ 1 ] )
                 {
                    pheno_covar_snp = false;
                 }
                 else
                 {
                    in_file.close();
                    std::cout << "error : read_configfile() : valid values for pheno_covar_snp are "
                              << "true and false. Invalid string " << line_elements[ 1 ] 
                              << " found in " << config_filename << std::endl;  
                    return FILE_ERROR;
                 }
                 found_pheno_covar_snp = true;
              } //if "pheno_covar_snp"
              else
              if ( "out_freq" == line_elements[0] )
              {
                  try
                  {
                     out_freq = std::stoull(line_elements[1]);
                  }
                  catch( std::exception & e)
                  {
                      in_file.close();
                      std::cout << "error : read_configfile() : error parsing string " 
                                << line_elements[1]
                                << " for out_freq in " << config_filename << std::endl;
                      return FILE_ERROR;
                  }
              } // out_freq
              else
            if ( "min_call_rate" == line_elements[0] )
            {
              try
              {
                 min_call_rate = std::stod(line_elements[1]);
              }
              catch( std::exception & e)
              {
                in_file.close();
                std::cout << "Error parsing string " << line_elements[1]
                      << " for min_call_rate in " << config_filename << std::endl;
                return FILE_ERROR;
              }
              if ( ( min_call_rate > 1.0 ) ||
                   ( min_call_rate < 0.0 ) )
              {
                in_file.close();
                std::cout << "Invalid value " << line_elements[1]
                      << " for min_call_rate in " << config_filename << std::endl;
                return FILE_ERROR;
              }
            } //min_call_rate
              else
              if ( "write_covariance" == line_elements[ 0 ] )
              {
                 if ( "true" == line_elements[ 1 ] )
                 {
                    write_covariance = true;
                 }
                 else
                 if ( "false" == line_elements[ 1 ] )
                 {
                    write_covariance = false;
                 }
                 else
                 {
                    in_file.close();
                    std::cout << "error : read_configfile() : valid values for write_covariance are "
                              << "true and false. Invalid string " << line_elements[ 1 ] 
                              << " found in " << config_filename << std::endl;  
                    return FILE_ERROR;
                 }
              } //if "write_covariance"
              else
              if ( "write_full_estimate" == line_elements[ 0 ] )
              {
                 if ( "true" == line_elements[ 1 ] )
                 {
                    write_full_estimate = true;
                 }
                 else
                 if ( "false" == line_elements[ 1 ] )
                 {
                    write_full_estimate = false;
                 }
                 else
                 {
                    in_file.close();
                    std::cout << "error : read_configfile() : valid values for write_full_estimate are "
                              << "true and false. Invalid string " << line_elements[ 1 ]
                              << " found in " << config_filename << std::endl;
                    return FILE_ERROR;
                 }
              } //if "write_full_estimate"
          } //while
           in_file.close();  
           const bool REQUIRED_PARAM_MISSING = 
                 (  ( false == found_geno_filename ) ||
                    ( false == found_pheno_filename ) ||
                    ( false == found_pheno_covar_snp ) );
           if ( true == REQUIRED_PARAM_MISSING )
           {
               if ( false == found_geno_filename )
               {
                  std::cout << "error : read_configfile() : geno_filename not found in "
                        << config_filename << std::endl;
               }
               if ( false == found_pheno_filename )
               {
                  std::cout << "error : read_configfile() : pheno_filename "
                        << "not found in " << config_filename << std::endl;
               }
               if ( false == found_pheno_covar_snp )
               {
                  std::cout << "error : read_configfile() : pheno_covar_snp not found in "
                        << config_filename << std::endl;
                  std::cout << "Valid settings for " << 
                        "pheno_covar_snp are the literal strings true and false." << 
                            std::endl;          
               }
              return FILE_ERROR;
           }
           return NO_ERROR;
     } //read_configfile

    int read_geno( const std::string & filename, 
                   std::vector< VectorXd > & snp_vxd_vec,
                   std::vector<std::string> & snp_name_vector,
                   const std::vector<subject_data_type> & 
                                        pheno_subject_data_vec,
                   double min_call_rate
                )
    {
        if ( ( min_call_rate < 0.0 ) || ( min_call_rate > 1.0 ) )
        {
           std::cout << 
         "error: read_geno() : min_call_rate should be between 0.0 and 1.0" 
                     << std::endl << "min_call_rate == " 
                     << min_call_rate << std::endl;
           return ARG_ERROR;
        }
        std::string line;
        std::ifstream in_file( filename.c_str() );
        if (!in_file.is_open())
        {
            std::cout << "read_geno : ";
            std::cout << 
                "error : could not open " << filename << std::endl;
            return FILE_ERROR;
        }
        const int NA_geno_value = -10; //should be negative
        snp_vxd_vec.clear();
        snp_name_vector.clear();
        while (getline(in_file,line))
        {
            VectorXd snp_vxd;
            std::string buff;
            std::stringstream ss(line);
            std::vector <double> line_elements;
            std::deque <std::string> line_strings;
            while (ss >> buff)
            {
                line_strings.push_back( buff );
            }
            if ( line_strings.size() > 1 )
            {
                snp_name_vector.push_back( line_strings[ 0 ] );
                line_strings.pop_front( );
            }
            else
            {
                in_file.close();
                std::cout << 
                   "error : read_geno() : each line must have at least 2 "
                          << "columns" << std::endl;  
                return FILE_ERROR; 
            }
            for ( const auto & mystr : line_strings )
            {
                int current_int;
                try {
                      current_int = stoi( mystr );
                      if ( ( current_int != 0 ) &&
                           ( current_int != 1 ) &&
                           ( current_int != 2 ) )
                        {     
                            std::cout <<
                            "read_geno() : error : invalid integer "
                               << current_int << " read from " << 
                                       filename << std::endl;
                            std::cout << 
                              "Valid interger values are 0, 1, and 2. " << 
                              "NA is allowed for missing data." << 
                              std::endl;
                            in_file.close();
                            return FILE_ERROR;        
                        } //if
                        line_elements.push_back( current_int ); 
                     }
                    catch( std::exception & e )
                    {
                        if ( mystr.compare( "NA" ) )
                        {
                            std::cout <<
                  "read_geno() : error : expected integer or NA but got "
                          << mystr << " when reading SNP data from " << 
                          filename << std::endl;
                            in_file.close();
                            return FILE_ERROR;
                        } //if
                        line_elements.push_back( NA_geno_value );
                    } //catch
            } //for
            snp_vxd = VectorXd::Zero( line_elements.size() );
            for (size_t index = 0; index < line_elements.size(); index++)
            {
                snp_vxd( index ) = line_elements[ index ];                     
            }
            snp_vxd_vec.push_back( snp_vxd );
        }   //while
        in_file.close();
        std::vector< VectorXd > passed_vxd_vec;
        std::vector< std::string > passed_snp_names_vec;
        size_t current_mat_index = 0;
        for ( auto & current_mat : snp_vxd_vec )
        {
           bool call_rate_check_passed;
           size_t missing_count = 0;
           for ( size_t my_case = 0; my_case < current_mat.cols(); 
                                   ++my_case )
           {
               if  ( current_mat( my_case ) < ( NA_geno_value / 2.0 ) )
               {
                   missing_count++;
               } //if
           } //for case
           const size_t MAX_ALLOWED_MISSING = 
                    round( ( 1.0 - min_call_rate ) * current_mat.cols() );
           if ( missing_count > (MAX_ALLOWED_MISSING ) )
           {
               call_rate_check_passed = false;
           }
           else
           {
               call_rate_check_passed = true;
           }
           int my_int;
           bool all_equal = true;
           for ( size_t  my_case = 0; my_case < current_mat.rows(); 
                                    ++my_case )            
           {
             if ( my_case == 0 )
             {
                my_int = int( current_mat( my_case ) );
             }                   
             else
             {
               if ( my_int != int( current_mat( my_case ) ) )
               {
                   all_equal = false;
                   break;
               }                         
             }     
          }
          if ( all_equal == true )
          {
              call_rate_check_passed = false;
          }
         if ( call_rate_check_passed )
         { 
           passed_vxd_vec.push_back( current_mat );
           passed_snp_names_vec.push_back( 
                   snp_name_vector[ current_mat_index ] );
         }
         current_mat_index++;
      } //for matrix
        snp_vxd_vec = passed_vxd_vec;
        snp_name_vector = passed_snp_names_vec;
        //impute missing if possible
        //if there are no subjects with non-missing data for a snp, 
        //remove snp        
        size_t snp_index = 0;
        std::vector< VectorXd > valid_vxd_vec;
        std::vector< std::string > valid_names_vec;
        for ( auto & my_snp : snp_vxd_vec )
        {
           size_t N_subjects_with_geno = 0;
           size_t S_subjects_with_geno = 0;
           for ( size_t my_subject = 0; my_subject < my_snp.rows(); ++my_subject )
           {
      
               if ( ( my_snp( my_subject) ) > (NA_geno_value / 2.0 ) )
               {
                  N_subjects_with_geno++;
                  S_subjects_with_geno += (size_t) my_snp( my_subject );
               } //if 
           } //for my_subject
           double mean_geno = 0.0;
           if ( N_subjects_with_geno > 0 )
           { 
                 mean_geno = round( (1.0 * S_subjects_with_geno) 
                                          / N_subjects_with_geno );
           }   
           bool valid_snp = false;
           if ( N_subjects_with_geno == my_snp.rows() )
           {
              valid_snp = true;
           }
           else
           if ( N_subjects_with_geno > 0 )
           {
               for ( size_t my_subject = 0; my_subject < my_snp.rows(); 
                                                 ++my_subject )
               {
                   if ( ( my_snp( my_subject) ) < (NA_geno_value / 2.0 ) )
                   {
                      my_snp( my_subject ) = mean_geno;   
                   } //if
                } //for my_subject
                valid_snp = true;
           } //if else
           if ( true == valid_snp )
           {
              valid_vxd_vec.push_back( my_snp );
              valid_names_vec.push_back( snp_name_vector[ snp_index ] );
           }
           snp_index++;
        } //for my_snp
        snp_vxd_vec = valid_vxd_vec;
        snp_name_vector = valid_names_vec;
        std::vector< std::vector< double > > snp_d_vectors;
        
        for ( const auto & my_snp : snp_vxd_vec )
        {
           std::vector< double > d_vector;
           for ( size_t subject = 0; subject < my_snp.rows(); ++subject )
           {
              if ( pheno_subject_data_vec[ subject ] == FULL_DATA )
              {
                 d_vector.push_back( my_snp( subject ) );
              }
           }
           snp_d_vectors.push_back( d_vector );
        } //for my_snp
        std::vector< VectorXd > snps_for_subjects_with_pheno;
        for ( size_t index = 0; index < snp_vxd_vec.size(); ++index )
        {
            const std::vector< double > my_d_vec = snp_d_vectors[ index ];
            const size_t PHENO_N = my_d_vec.size();
            VectorXd subjects_with_pheno_vxd = VectorXd::Zero( PHENO_N );
            for ( size_t subject = 0; subject < PHENO_N; ++subject )
            {
                subjects_with_pheno_vxd( subject ) = my_d_vec[ subject ]; 
            }
            snps_for_subjects_with_pheno.push_back( subjects_with_pheno_vxd );
        }
        snp_vxd_vec = snps_for_subjects_with_pheno;
        return NO_ERROR;
    } //method

    int make_mxd_with_constant_from_covar_mxd( const MatrixXd & covar_mxd,
                                       MatrixXd & covar_with_constant_mxd )
    {
        const size_t NUM_SUBJECTS = covar_mxd.cols();
        const size_t NUM_COVAR = covar_mxd.rows();
        covar_with_constant_mxd = MatrixXd::Zero( NUM_COVAR + 1, 
                                                  NUM_SUBJECTS );
        for ( size_t col = 0; col < NUM_SUBJECTS; col++)
        {
            covar_with_constant_mxd( 0, col ) = 1.0;
        }
        covar_with_constant_mxd.block( 1, 0, NUM_COVAR, NUM_SUBJECTS ) =
            covar_mxd;
        return NO_ERROR;
    } 
    int make_mxd_with_constant_and_snp_from_covar_mxd( 
                    const MatrixXd & covar_mxd,
                    const VectorXd & snp_vxd,
                    MatrixXd & covar_with_constant_and_snp_mxd )
    {
        const size_t NUM_SUBJECTS = covar_mxd.cols();
        const size_t NUM_COVAR = covar_mxd.rows();
        if ( snp_vxd.rows() != NUM_SUBJECTS )
        {
            std::cout << 
              "Error: make_mxd_with_constant_and_snp_from_covar_mxd: " <<
                "snp_vxd.rows() != covar_mxd.cols() " << std::endl;
            return DIM_ERROR;
        }
        
        covar_with_constant_and_snp_mxd = MatrixXd::Zero( NUM_COVAR + 2, 
                                                          NUM_SUBJECTS );
        for ( size_t col = 0; col < NUM_SUBJECTS; col++)
        {
           covar_with_constant_and_snp_mxd( 0, col ) = 1.0;
           covar_with_constant_and_snp_mxd( 1, col ) = snp_vxd( col );
        }
        covar_with_constant_and_snp_mxd.block( 2, 0, NUM_COVAR, 
                                                     NUM_SUBJECTS ) =
            covar_mxd;
        
        return NO_ERROR;
    }

    void calc_z_stat_vec( const VectorXd & theta_vec,
                      const VectorXd & se_col_vec,
                            VectorXd & z_stat_vec )
    {
        const size_t NUM_COOR  = theta_vec.rows(); //p
        z_stat_vec = VectorXd::Zero(NUM_COOR );
        for (size_t index = 0; index < NUM_COOR; index++)
        {
            z_stat_vec( index ) =
                theta_vec( index ) / se_col_vec( index );
        }
    }

    void calc_p_val_vec( const VectorXd & z_stat_vec, VectorXd & p_val_vec )
    {
        // chi squared       
        const int DOF = 1;               
        const size_t NUM_COOR  = z_stat_vec.rows(); //p
        p_val_vec = VectorXd::Zero(NUM_COOR);
        for (size_t index = 0; index < NUM_COOR; index++)
        {
            p_val_vec( index )  =
                boost::math::gamma_q( DOF /2.0 ,     
                                ( z_stat_vec(index) *
                                  z_stat_vec(index) ) / 2.0);
        }
    }

    double calc_variance( const std::vector<double> & double_vector )
    {
        double rv = 0.0;
        double sum = 0.0;
        const size_t SIZE = double_vector.size();
        if ( ! SIZE )
        {
            const double default_val = 0.0;
            std::cout << "calc_variance() : SIZE == 0" <<
                std::endl;
            std::cout << "returning default == " << default_val <<
                std::endl;
            return default_val;
        }
        for ( size_t index = 0; index < SIZE;
                ++index )
        {
            sum += double_vector[ index ];
        }
        const double mean = sum / SIZE;
        sum = 0.0;
        for ( size_t index = 0; index < SIZE;
                ++index )
        {
            const double diff = mean - double_vector[ index ];
            const double sqr_diff = diff * diff;
            sum += sqr_diff;
        }
        rv = sum / SIZE;
        return rv;
    } //calc_variance

    bool isVectorConverged( const VectorXd & old_vec, const VectorXd & new_vec,
                            double convergence_ratio )
    {
        if ( old_vec.rows() != new_vec.rows() )
        {
            std::cout << "isVectorConverged() : old_vec and new_vec have different lengths." << std::endl;
            return false;
        }
        double max_delta = 0;
        for ( size_t index = 0; index < old_vec.rows(); ++index )
        {
            const double old_val = old_vec( index );
            const double new_val = new_vec( index );
            const double fabs_diff =
                fabs( new_val - old_val );
            double delta_val;
            if ( fabs( old_val ) >= 0.01 )
            {
                delta_val  = fabs_diff / fabs( old_val );  
            }
            else
            {
                delta_val  = fabs_diff;
            }
            
            if ( delta_val  > max_delta )
            {
                max_delta = delta_val;
                if ( max_delta > convergence_ratio )
                {
                    return false;
                }
            } //if
        } //for
        return true;
    } //isVectorConverged

    void int_to_error_string( int myint, std::string & my_error_str )
    {
         switch( myint )
        {
            case NO_ERROR:
                 my_error_str="NO_ERROR";
                 break;
            case DIM_ERROR:
                 my_error_str="DIM_ERROR";
                 break;
            case ARG_ERROR:
                 my_error_str="ARG_ERROR";
                 break;
            case NOT_CONVERGED_ERROR:
                 my_error_str="NOT_CONVERGED_ERROR";
                 break;
            case FILE_ERROR:
                 my_error_str="FILE_ERROR";
                 break;
            case INVERT_ERROR:
                 my_error_str="INVERT_ERROR";
                 break;
            default:
                my_error_str="UNKNOWN ERROR CODE";
        }
    }
} //namespace
